/********************************************************************************
*                                                                               *
*                     T h e   A d i e   T e x t   E d i t o r                   *
*                                                                               *
*********************************************************************************
* Copyright (C) 1998,2003 by Jeroen van der Zijp.   All Rights Reserved.        *
*********************************************************************************
* This program is free software; you can redistribute it and/or modify          *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation; either version 2 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* This program is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with this program; if not, write to the Free Software                   *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
*********************************************************************************
* $Id: Adie.h,v 1.39 2003/02/10 06:09:07 fox Exp $                              *
********************************************************************************/
#ifndef ADIE_H
#define ADIE_H


// Version
#define VERSION_MAJOR 1
#define VERSION_MINOR 3
#define VERSION_PATCH 1


class HelpWindow;
class Preferences;
class TextWindow;


// Main Application class
class Adie : public FXApp {
  FXDECLARE(Adie)
  friend class TextWindow;
protected:
  TextWindowList  windowlist;           // Window list
  FXRecentFiles   mrufiles;             // Recent files list
  FXFileDict     *associations;         // File association table
  FXIcon         *bigicon;              // Big application icon
  FXIcon         *smallicon;            // Small application icon
  FXIcon         *newicon;
  FXIcon         *reloadicon;
  FXIcon         *openicon;
  FXIcon         *saveicon;
  FXIcon         *saveasicon;
  FXIcon         *printicon;
  FXIcon         *cuticon;
  FXIcon         *copyicon;
  FXIcon         *pasteicon;
  FXIcon         *deleteicon;
  FXIcon         *undoicon;
  FXIcon         *redoicon;
  FXIcon         *fontsicon;
  FXIcon         *helpicon;
  FXIcon         *quiticon;
  FXIcon         *searchicon;
  FXIcon         *searchnexticon;
  FXIcon         *searchprevicon;
  FXIcon         *bookseticon;
  FXIcon         *booknexticon;
  FXIcon         *bookprevicon;
  FXIcon         *bookdelicon;
  FXIcon         *shiftlefticon;
  FXIcon         *shiftrighticon;
private:
  Adie(){}
  Adie(const Adie&);
  Adie& operator=(const Adie&);
  TextWindow *findUntitled() const;
  TextWindow *findWindow(const FXString& file) const;
  FXString Adie::uniqueName() const;
public:
  static const FXchar untitled[];
public:
  enum{
    ID_NEW=FXApp::ID_LAST,
    ID_OPEN,
    ID_OPEN_RECENT,
    ID_OPEN_SELECTED,
    ID_CLOSE,
    ID_WINDOW_1,
    ID_WINDOW_2,
    ID_WINDOW_3,
    ID_WINDOW_4,
    ID_WINDOW_5,
    ID_WINDOW_6,
    ID_WINDOW_7,
    ID_WINDOW_8,
    ID_WINDOW_9,
    ID_WINDOW_10,
    ID_LAST
    };
public:
  long onCmdNew(FXObject*,FXSelector,void*);
  long onCmdOpen(FXObject*,FXSelector,void*);
  long onCmdOpenRecent(FXObject*,FXSelector,void*);
  long onCmdOpenSelected(FXObject*,FXSelector,void*);
  long onCmdClose(FXObject*,FXSelector,void*);
  long onUpdTitle(FXObject*,FXSelector,void*);
  long onCmdQuit(FXObject*,FXSelector,void*);
  long onCmdWindowSelect(FXObject*,FXSelector,void*);
  long onUpdWindowSelect(FXObject*,FXSelector,void*);
public:

  // Construct application object
  Adie(const FXString& name);

  // Initialize application
  virtual void init(int& argc,char** argv,FXbool connect=TRUE);

  // Return TRUE if windows have been loaded
  FXbool hasWindows() const { return windowlist.no()>0; }

  // Open new window x
  TextWindow* newWindow(const FXString& file);

  // Open window to existing file
  TextWindow *openWindow(const FXString& file);

  // Close window
  void closeWindow(TextWindow *window);

  // Close all windows
  void closeAllWindows();

  // Delete application object
  virtual ~Adie();
  };

#endif

